//
//	SolitariumLib - Copyright  2005, glenn andreas, All Rights Reserved
//

// Define default card size parameters
var slCardWidth = 20;
var slCardHeight = 35;
var slUpCardsYOverlap = 10; // how many pixels we need to show when covering face up
var slDownCardsYOverlap = 7; // how many pixels we need to show when covering face up
var slUpCardsXOverlap = 0; // how many pixels we need to show when covering face up
var slDownCardsXOverlap = 0; // how many pixels we need to show when covering face up

var slNumDecks = 1;

var slCheatPeek = false;
var slCardTable = null; // will be set in init
var slCardTableWidth = 0;
var slCardTableHeight = 0;
var slCardSkin = ""; // to support card skins
var slCardSize = "Large";
var slCardBack = "Default"; // default card back value
var slCardBackSubDir = "Backs/";
var slCardDesign = "Default";
var slCardDesignSubDir = "Cards/"; // what subdir the cards are stored in (include path sep)
var slCardMarkersSubDir = "Markers/";
var slCardImgExt = ".png"; // stored as pngs

var slWinTimer = null;
var slFlyTimer = null;
var slFlyStartX = null;
var slFlyStartY = null;
var slFlyQueue = new Array();
var slFlyEnabled = false;

// Card value is the number followed by suit, where number is A, 1, 2, ..., 9, T, J, Q, K
// and suit is H C S D
function slCardSuit(cardValue)
{
	return cardValue.slice(1,2);
}

function slCardPip(cardValue)
{
	var pip = cardValue.slice(0,1);
	switch (pip) {
		case 'A': return 1;
		case 'T': return 10;
		case 'J': return 11;
		case 'Q': return 12;
		case 'K': return 13;
	}
	return parseInt(pip)
}

function slCardColor(cardValue)
{
	var suit = slCardSuit(cardValue);
	return suit == 'H' || suit == 'D';
}

function slCardName(cardValue)
{
	var suitNames = {'H':'Hearts','C':'Clubs','D':'Diamonds','S':'Spades'};
	var pipNames = new Array('Joker','Ace','Two','Three','Four','Five','Six','Seven','Eight','Nine','Ten','Jack','Queen','King');
	return	pipNames[slCardPip(cardValue)] + " of " + suitNames[slCardSuit(cardValue)];
}
function slCardImage(cardValue)
{
	// pass null to get the back side
	if (cardValue == null) {
		return "url(file:" + slCardBackSubDir + slCardSize + "/" + slCardBack + slCardImgExt+ ")";
	} else {
		return "url(file:" + slCardDesignSubDir + slCardDesign + "/" + slCardSize + "/" + cardValue + slCardImgExt+ ")";
	}
}
function slMarkerImage(marker)
{
	return "url(file:" + slCardMarkersSubDir + slCardSize + "/" + marker + slCardImgExt+ ")";
}

var slDoubleClickLast = null;
var slDoubleClickWhat = null;
function slCardClicked(event)
{
	if (slWinTimer) return false; // do nothing while in win animation
	if (slFlyTimer) return false; // do nothing while flying
	if (slDoubleClickLast) {
		if ((new Date()).getTime() - slDoubleClickLast.getTime() < 500 && slDoubleClickWhat == event.target) {
			slDoubleClickLast = null;
			return event.target.pile.doubleClickRoutine(event.target);
		}
	}
	slDoubleClickLast = new Date();
	slDoubleClickWhat = event.target;
	return event.target.pile.clickRoutine(event.target, event.clientX, event.clientY)
}
function slCardCursor(event)
{
	if (slWinTimer) {
		event.target.style.cursor = null; // do nothing while flying
		return;
	}
	if (slFlyTimer) {
		event.target.style.cursor = "wait"; // do nothing while flying
		return;
	}
	event.target.style.cursor = event.target.pile.cursorRoutine(event.target);
}

function slBuildCard(x, y, cardValue, faceUp)
{
	var card = document.createElement('div');
	card.setAttribute('class','slCard');
	card.style.position = 'absolute';
	card.style.left = x + 'px';
	card.style.top = y + 'px';
	card.style.width = slCardWidth + 'px';
	card.style.height = slCardHeight + 'px';
	card.cardValue = cardValue;

	card.faceUp = faceUp;
	if (card.faceUp) {
		card.style.background = slCardImage(cardValue);
		card.title = slCardName(cardValue);
	} else {
		card.style.background = slCardImage(null);
		card.title = null;
	}
	return card;
}

function slRebrandCard(cardDiv)
{
	if (cardDiv.faceUp) {
		cardDiv.style.background = slCardImage(cardDiv.cardValue);
		cardDiv.title = slCardName(cardDiv.cardValue);
	} else {
		cardDiv.style.background = slCardImage(null);
		cardDiv.title = null;
	}
}

function slFlipCardUp(cardDiv)
{
	if (!cardDiv.faceUp) {
		cardDiv.faceUp = true;
		cardDiv.style.background = slCardImage(cardDiv.cardValue);
		cardDiv.title = slCardName(cardDiv.cardValue);
	}
}
function slFlipCardDown(cardDiv)
{
	if (cardDiv.faceUp) {
		cardDiv.faceUp = false;
		cardDiv.style.background = slCardImage(null);
		cardDiv.title = null;
	}
}

var slShuffledDeck = null;

function slShuffleArray(deck) // shuffle an array - usually a deck
{
	var seed = ((new Date()).getTime() % 1000) / 1000; // make seed to offset math.random
	for (var i=0;i<deck.length;i++) {
		// go through and swap pairs of cards
		var i1 = Math.floor(((Math.random() + seed) % 1.0) * deck.length);
		var i2 = Math.floor(((Math.random() + seed) % 1.0) * deck.length);
		var t = deck[i1];
		deck[i1] = deck[i2];
		deck[i2] = t;
	}
}

function slShuffleDeck()
{
	slShuffledDeck = new Array();
	for (var i=0;i<slNumDecks;i++) {
		for (var suiti = 0; suiti < 4; suiti++) {
			var suit = "HCSD".slice(suiti,suiti+1);
			for (var pipi = 0; pipi < 13; pipi++) {
				var pip = "A23456789TJQK".slice(pipi,pipi+1);
				slShuffledDeck.push(pip + suit);
			}
		}
	}
	slShuffleArray(slShuffledDeck); // and shuffle it up
}

function slRemoveFromDeck(cards)
{
	for (var i=0;i<slShuffledDeck.length; i++) {
		for (var j=0;j<cards.length;j++) {
			if (slShuffledDeck[i] == cards[j]) {
				// remove it
				slShuffledDeck.splice(i,1);
				i--;
			}
		}
	}
}
function slNextCardFromDeck()
{
	if (slShuffledDeck.length) {
		return slShuffledDeck.pop();
	}
	return null; // no more cards
}
function slPlaceOnBottomOfDeck(card)
{
	// put it back on the bottom of the deck
	slShuffledDeck.splice(0,0,card);
}

function slPileAtXY(x,y, cardDiv)
{
	for (var i = 0; i<slAllPiles.length; i++) {
		if (cardDiv.pile == slAllPiles[i]) {
			continue; // ignore dropping on starting pile
		}
		if (slAllPiles[i].hitTestDrop(x,y)) {
			return slAllPiles[i];
		}
	}
	return null;
}

var slCurrentDragList = null;
var slDragDeltaX;
var slDragDeltaY;
var slDropPile = null;

function slDragEnd()
{
	document.onmousemove = null;
	document.onmouseup = null;
	if (slDropPile) {
		slDropPile.unhighlightDrop();
		//alert('will drop at ' + slDropPile.div.id);
		for (var i =slCurrentDragList.length-1; i>=0; i--) {
			slCurrentDragList[i].pile.removeTop(); // remove this old card (going backwards)
		}
		for (var i =0;i<slCurrentDragList.length;i++) { // add them one at a time
			slDropPile.addToTop(slCurrentDragList[i]);
			slDropPile.postAddToTop();
		}
		slCurrentDragList = null;
		slDropPile = null;
		slCheckForWin();
	} else {
		// drag back to start
		for (var i=0;i<slCurrentDragList.length;i++) {
			slCurrentDragList[i].style.zIndex = slCurrentDragList[i].origIndex;
			slCurrentDragList[i].style.left = slCurrentDragList[i].origX;
			slCurrentDragList[i].style.top = slCurrentDragList[i].origY;
		}
		slCurrentDragList = null;
	}
}

sl_lastPile = null;
function slDragDrag(event)
{
	slDoubleClickLast = null; // disable double click stuff
	var x = event.clientX;
	var y = event.clientY;
	for (var i=0;i<slCurrentDragList.length;i++) {
		var cx = x - slCurrentDragList[i].lastMouseX;
		var cy = y - slCurrentDragList[i].lastMouseY;
		// convert from pile to board space
		var pileDiv = slCurrentDragList[i].offsetParent;
		cx += pileDiv.offsetLeft;
		cy += pileDiv.offsetTop;
		if (cx < 0) {
			x += -cx;
		}
		if (cx + slCardWidth > slCardTableWidth) {
			x -= cx - (slCardTableWidth - slCardWidth);
		}
		if (cy < 0) {
			y += -cy;
		}
		if (cy > slCardTableHeight) {
			// allow to go over the bottom a little
			y -= cy - slCardTableHeight;
		}
	}
	// should really clip it here
	for (var i=0;i<slCurrentDragList.length;i++) {
		slCurrentDragList[i].style.left = x - slCurrentDragList[i].lastMouseX;
		slCurrentDragList[i].style.top = y - slCurrentDragList[i].lastMouseY;
	}
	// figure out 'track point' (convert to slboard coord) based on top middle of bottom card
	var div = slCurrentDragList[0].offsetParent;
	x = slCurrentDragList[0].offsetLeft + slCurrentDragList[0].offsetWidth / 2;
	y = slCurrentDragList[0].offsetTop;
	while (div != slCardTable) {
		x += div.offsetLeft;
		y += div.offsetTop;
		div = div.offsetParent;
	}
	var pile = slPileAtXY(x,y, slCurrentDragList[0]);
	//if (pile != sl_lastPile) {
	//	if (pile) alert('over ' + pile.div.id);
	//	else alert('over null');
	//	sl_lastPile = pile;
	//}
	if (pile) {
		//alert('try drop at ' + pile.div.id);
		var cantDrag = false;
		if (pile.cantDragFrom) {
			for (var j=0;j < pile.cantDragFrom.length;j++) {
				if (slCurrentDragList[0].pile == pile.cantDragFrom[j]) {
					cantDrag = true;
					break;
				}
			}
		}
		if (pile == slDropPile || cantDrag) {
			// do nothing - already dropping here
		} else if (pile.canDropCardList(slCurrentDragList)) {
			if (slDropPile) slDropPile.unhighlightDrop();
			slDropPile = pile;
			slDropPile.highlightDrop();
			//alert('can drop at ' + pile.div.id);
		}
	} else {
		if (slDropPile) slDropPile.unhighlightDrop();
		slDropPile = null;
	}
	return false;
}

function slStartDrag(cardDivList, x, y)
{
	//alert('Starting drag of ' + cardDiv.cardValue);
	for (var i=0;i<cardDivList.length;i++) {
		var cardDiv = cardDivList[i];
		cardDiv.origX = cardDiv.style.left;
		cardDiv.origY = cardDiv.style.top;
		cardDiv.origIndex = cardDiv.style.zIndex;
		cardDiv.style.zIndex = 1000 + i; // drag on top of everything
		cardDiv.lastMouseX = x - parseInt(cardDiv.style.left);
		cardDiv.lastMouseY = y - parseInt(cardDiv.style.top);
	}
	document.onmousemove = slDragDrag;
	document.onmouseup = slDragEnd;
	slCurrentDragList = cardDivList;
	return false;
}


function slFlyNow()
{
	var fly = slFlyQueue[0];
	var dx = fly['endX'] - fly['card'].offsetLeft;
	var dy = fly['endY'] - fly['card'].offsetTop;
	if (dx * dx + dy * dy <= (slCardWidth * .75) * (slCardWidth * .75)) {
		// less than half a card width away - done
		fly['card'].style.left = fly['endX'];
		fly['card'].style.top = fly['endY'];
		fly['card'].style.zIndex = fly['zindex']; // done with fly
		slFlyQueue.shift(); // this is done
		if (slFlyQueue.length == 0) { // no more to fly
			clearInterval(slFlyTimer);
			slFlyTimer = null;
		}
		return;
	}
	var flyDist = (slCardWidth / 2);
	var angle = Math.atan2(fly['endX'] - fly['startX'], fly['endY'] - fly['startY']);
	dx = flyDist * Math.sin(angle);
	dy = flyDist * Math.cos(angle);
	fly['card'].style.left = fly['card'].offsetLeft + dx;
	fly['card'].style.top = fly['card'].offsetTop + dy;
}

function slFly(cardDiv, destPileDiv)
{
	if (!slFlyEnabled) {
		slFlyStartX = null;
		return;
	}
	if (slFlyStartX == null) return; // don't fly
	slFlyQueue.push({'startX': slFlyStartX - destPileDiv.offsetLeft, //convert from table to pile coordinates
		'startY': slFlyStartY - destPileDiv.offsetTop,
		'endX' : cardDiv.offsetLeft,
		'endY' : cardDiv.offsetTop,
		'zindex' : cardDiv.style.zIndex,
		'card' : cardDiv
	});
	// move to start point
	cardDiv.style.left = slFlyStartX - destPileDiv.offsetLeft;
	cardDiv.style.top = slFlyStartY - destPileDiv.offsetTop;
	cardDiv.style.zIndex = 950 - slFlyQueue.length; // bring up to fly
	slFlyStartX = null;
	slFlyStartY = null;
	if (slFlyTimer == null) { // start the timer
		slFlyTimer = setInterval("slFlyNow();",50);
	}
}
var sl_lastTest = null;
function sl_logRule(topCard, dropArray)
{
	return;
	if (topCard + dropArray[0] != sl_lastTest) {
		sl_lastTest = topCard + dropArray[0];
		alert('rule:' + dropArray[0] + "(" + slCardColor(dropArray[0]) + ',' + slCardPip(dropArray[0]) + ')' +
		' on ' + topCard+ "(" + slCardColor(topCard) + ',' + slCardPip(topCard) + ')');
	}
}
// various drop rules
slBlankAceDesc = "Spaces can only be filled with an Ace.";
slBlankKingDesc = "Spaces can only be filled with an King.";
slBlankAnyDesc = "Spaces can be filled with any card.";
slBlankNoneDesc = "Spaces remain empty once the last card is removed.";

function slDropRuleFoundation(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// for play on foundation, only one card
	if (dropArray.length != 1) return false;
	// unsuited, blank pile, put any ace here
	if (topCard == "0X") return slCardPip(dropArray[0]) == 1;
	if (topCard == "*X") return true;
	// match suit
	if (slCardSuit(topCard) != slCardSuit(dropArray[0])) return false;
	// go up by one
	if (slCardPip(topCard)+1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleFoundation.ruleDesc = "Build up from Ace to King, in suit, one card at a time.";
slDropRuleFoundation.blankDesc = slBlankAceDesc;

function slDropRuleUnsuitedFoundation(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// for play on foundation, only one card
	if (dropArray.length != 1) return false;
	// unsuited, blank pile, put any ace here
	if (topCard == "0X") return slCardPip(dropArray[0]) == 1;
	if (topCard == "*X") return true;
	// no need to match suit
	//if (slCardSuit(topCard) != slCardSuit(dropArray[0])) return false;
	// go up by one
	if (slCardPip(topCard)+1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleUnsuitedFoundation.ruleDesc = "Build up from Ace to King, regardless of suit, one card at a time.";
slDropRuleUnsuitedFoundation.blankDesc = slBlankAceDesc;

var slGlobalFoundationBase = null;
function slDropRuleUnbasedFoundation(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// for play on foundation, only one card
	if (dropArray.length != 1) return false;
	// unsuited, blank pile, put any ace here
	if (topCard == "0X" || topCard == "*X") {
		if (slGlobalFoundationBase == true) {
			return true; // we don't care, use what we've got
		} else if (slGlobalFoundationBase) {
			// foundation base set - share it
			return slGlobalFoundationBase == slCardPip(dropArray[0])
		} else {
			// set the foundation to this
			slGlobalFoundationBase = slCardPip(dropArray[0]);
			return true;
		}
	}
	// match suit
	if (slCardSuit(topCard) != slCardSuit(dropArray[0])) return false;
	// go up by one
	if ((slCardPip(topCard)+1) % 13 != slCardPip(dropArray[0]) % 13) return false;
	return true;
}
slDropRuleUnbasedFoundation.ruleDesc = "Build a full 13 cards, one card at a time, wrapping from King to Ace if needed.";
slDropRuleUnbasedFoundation.blankDesc = "The first foundation filled determines where the other foundations start";

function slDropRuleAltColorDown(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	// alternate suit
	if (slCardColor(topCard) == slCardColor(dropArray[0])) return false;
	// go down by one
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleAltColorDown.description = "Build down in value, alternating color.";
slDropRuleAltColorDown.blankDesc = slBlankKingDesc;

function slDropRuleAltColorDownAnyBlank(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// can put any here
	if (topCard == "0X") return true;
	if (topCard == "*X") return true;
	// alternate suit
	if (slCardColor(topCard) == slCardColor(dropArray[0])) return false;
	// go down by one
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleAltColorDownAnyBlank.description = "Build down in rank, alternating color.";
slDropRuleAltColorDownAnyBlank.blankDesc = slBlankAnyDesc;

function slDropRuleSameColorDown(topCard, dropArray)
{
	sl_logRule(topCard,dropArray);
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	// alternate suit
	if (slCardColor(topCard) != slCardColor(dropArray[0])) return false;
	// go down by one
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleSameColorDown.description = "Build down in rank, mainting the same color.";
slDropRuleSameColorDown.blankDesc = slBlankKingDesc;

function slDropRuleSameSuitUpOrDown(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardSuit(topCard) != slCardSuit(dropArray[0])) return false;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0]) && slCardPip(topCard) + 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleSameSuitUpOrDown.description = "Build up or down in the same suit.";
slDropRuleSameSuitUpOrDown.blankDesc = slBlankKingDesc;

function slDropRuleSameSuitDown(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardSuit(topCard) != slCardSuit(dropArray[0])) return false;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropRuleSameSuitDown.description = "Build down in the same suit.";
slDropRuleSameSuitDown.blankDesc = slBlankKingDesc;

function slDropAnyUpOrDown(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0]) && slCardPip(topCard) + 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropAnyUpOrDown.description = "Build up or down, regardless of suit.";
slDropAnyUpOrDown.blankDesc = slBlankKingDesc;

function slDropAnyUpOrDownWrap(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardPip(topCard) == 1 && slCardPip(dropArray[0]) == 13) return true;
	if (slCardPip(topCard) == 13 && slCardPip(dropArray[0]) == 1) return true;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0]) && slCardPip(topCard) + 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropAnyUpOrDownWrap.description = "Build up or down, regardless of suit - aces can be played on kings, and kings on aces (wraps around).";
slDropAnyUpOrDownWrap.blankDesc = slBlankKingDesc;

function slDropAnyDownWrap(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardPip(topCard) == 1 && slCardPip(dropArray[0]) == 13) return true;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropAnyDownWrap.description = "Build down in rank, regardless of suit- aces can be played on kings, and kings on aces (wraps around).";
slDropAnyDownWrap.blankDesc = slBlankKingDesc;

function slDropAltDownWrap(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	// alternate color
	if (slCardColor(topCard) == slCardColor(dropArray[0])) return false;
	if (slCardPip(topCard) == 1 && slCardPip(dropArray[0]) == 13) return true;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropAnyDownWrap.description = "Build down in rank, regardless of suit- aces can be played on kings, and kings on aces (wraps around).";
slDropAnyDownWrap.blankDesc = slBlankKingDesc;

function slDropAnyDown(topCard, dropArray)
{
	// need a king to put in a blank spot
	if (topCard == "0X") return slCardPip(dropArray[0]) == 13;
	if (topCard == "*X") return true;
	if (slCardPip(topCard) - 1 != slCardPip(dropArray[0])) return false;
	return true;
}
slDropAnyDown.description = "Build down in rank, regardless of suit.";
slDropAnyDown.blankDesc = slBlankKingDesc;

function slDropRuleFullSuit(topCard, dropArray)
{
	// need a full suit to drop here
	if (topCard == "0X") {
		if (dropArray.length != 13) return false; // must be all 13
		for (var i=0;i<13;i++) {
			if (slCardPip(dropArray[i]) != 13-i) return false;
			if (slCardSuit(dropArray[i]) != slCardSuit(dropArray[0])) return false;
		}
		return true;
	}
	return false; // if anything here, can't drop
}
slDropRuleFullSuit.description = "Must drop the entire suit from King down to Ace at once.";
slDropRuleFullSuit.blankDesc = "";

function slDropRuleTotal14(topCard, dropArray)
{
	if (topCard == "0X") {
		return false; // can't drop any here
	}
	if (dropArray.length != 1) {
		return false; // only one card
	}
	if (slCardPip(dropArray[0]) + slCardPip(topCard) == 14) {
		return true;
	}
	return false; // if anything here, can't drop
}
slDropRuleTotal14.description = "Combine pairs of cards together to total 14.";
slDropRuleTotal14.blankDesc = "";
function slDropRuleTotal13(topCard, dropArray)
{
	if (topCard == "0X") {
		return false; // can't drop any here
	}
	if (dropArray.length != 1) {
		return false; // only one card
	}
	if (slCardPip(dropArray[0]) + slCardPip(topCard) == 13) {
		return true;
	}
	return false; // if anything here, can't drop
}
slDropRuleTotal14.description = "Combine pairs of cards together to total 13.";
slDropRuleTotal14.blankDesc = "";
function slDropRulePairs(topCard, dropArray)
{
	if (topCard == "0X") {
		return false; // can't drop any here
	}
	if (dropArray.length != 1) {
		return false; // only one card
	}
	if (slCardPip(dropArray[0]) == slCardPip(topCard)) {
		return true;
	}
	return false; // if anything here, can't drop
}
slDropRulePairs.description = "Combine pairs of cards of the same rank.";
slDropRulePairs.blankDesc = "";

function slDropKingsOnly(topCard, dropArray) // used for "discards kings"
{
	if (dropArray.length != 1) {
		return false; // only one card
	}
	return slCardPip(dropArray[0]) == 13;
}

function slDragRuleAltColorDown(cards)
{
	var lastCard = cards[0];
	for (var i=1;i<cards.length;i++) {
		var thisCard = cards[i];
		if (slCardColor(lastCard) == slCardColor(thisCard)) return false; // not alternating
		if (slCardPip(lastCard) - 1 != slCardPip(thisCard)) return false; // not decreasing
		lastCard = thisCard;
	}
	return true;// got through them all OK
}
slDragRuleAltColorDown.description = "alternating cards, in decreasing order";

function slDragRuleSameSuitUpOrDown(cards)
{
	var lastCard = cards[0];
	for (var i=1;i<cards.length;i++) {
		var thisCard = cards[i];
		if (slCardColor(lastCard) == slCardColor(thisCard)) return false; // not alternating
		if (slCardPip(lastCard) - 1 != slCardPip(thisCard) && slCardPip(lastCard) + 1 != slCardPip(thisCard)) return false; // not decreasing
		lastCard = thisCard;
	}
	return true;// got through them all OK
}
slDragRuleSameSuitUpOrDown.description = "cards of the same suit, in either increasing or decreasing order";

function slDragRuleSameSuitDown(cards)
{
	var lastCard = cards[0];
	for (var i=1;i<cards.length;i++) {
		var thisCard = cards[i];
		if (slCardSuit(lastCard) != slCardSuit(thisCard)) return false; // not same
		if (slCardPip(lastCard) - 1 != slCardPip(thisCard)) return false; // not decreasing
		lastCard = thisCard;
	}
	return true;// got through them all OK
}
slDragRuleSameSuitDown.description = "cards of the same suit, in decreasing order";

function slDragRuleSingleOrKingPile(cards)
{
	if (cards.length == 1) return true; // can do a single pile
	// or alternating cards, with the bottom being a king
	var lastCard = cards[0];
	for (var i=1;i<cards.length;i++) {
		var thisCard = cards[i];
		if (slCardColor(lastCard) == slCardColor(thisCard)) return false; // not alternating
		if (slCardPip(lastCard) - 1 != slCardPip(thisCard)) return false; // not decreasing
		lastCard = thisCard;
	}
	return slCardPip(lastCard) == 13;// got through them all OK, but this must be a king
}
slDragRuleSingleOrKingPile.description = "alternating cards in decreasing order if top is a King (or a single card)";

function slClickEmptyPile(event)
{
	//alert('click empty');
	return event.target.pile.emptyClickRoutine()
}

var slAllPiles = new Array();

function slFindBestDropPile(card)
{
	var best = null;
	for (var i=0;i<slAllPiles.length;i++) {
		if (slAllPiles[i] == card.pile) continue; // already there
		if (slAllPiles[i].cantDragFrom) {
			var cantDrag = false;
			for (var j=0;j < slAllPiles[i].cantDragFrom.length;j++) {
				if (card.pile == slAllPiles[i].cantDragFrom[j]) {
					cantDrag = true;
					break;
				}
			}
			if (cantDrag) continue;
		}
		if (slAllPiles[i].bestDropRank == 0) continue; // doesn't want auto double drop
		if (slAllPiles[i].isComplete()) continue;
		if (slAllPiles[i].canDropCardList(new Array(card))) {
			if (best == null) {
				best = slAllPiles[i];
			} else if (slAllPiles[i].bestDropRank > best.bestDropRank) {
				// otherwise a foundation beats the pile
				best = slAllPiles[i];
			}
		}
	}
	return best;
}

function slPileBase(div) {
	this.div = div;
	this.pile = new Array();
	slAllPiles.push(this); // keep a global list
	div.pile = this;
	this.cardXStagger = 0;
	this.cardYStagger = 0;
	this.baseZIndex = 0;
	this.canDragTop = false;
	this.dragPileRule = null; // for dragging multiple
	this.bestDropRank = 0; // to double click and drop
	this.pileLimit = 0;
	// build the empty image
	this.emptyImage = document.createElement('div');
	this.emptyImage.style.position = 'absolute';
	this.emptyImage.style.left = '0px';
	this.emptyImage.style.top = '0px';
	this.emptyImage.style.width = slCardWidth + 'px';
	this.emptyImage.style.height = slCardHeight + 'px';
	this.emptyImage.style.display = 'none';
	this.emptyImage.style.zIndex = 200; // above cards, below dragging
	this.emptyImage.onmousedown = slClickEmptyPile;
	this.emptyImage.pile = this;
	this.currentHighlight = null;
	this.div.appendChild(this.emptyImage);
	this.showsEmptyImage = false;
	this.emptyClickRoutine = function() {
		return false;
	}
	this.canDragTopRoutine = function(card) {
		return this.canDragTop;
	}
	this.doubleClickRoutine = function(card) {
		if (this.canDragTopRoutine() && card == this.topCard()) { // double click and drag top only
			var drop = slFindBestDropPile(card);
			if (drop) {
				// fly converted to table space
				var card = this.removeTop(true);
				drop.addToTop(card);
				drop.postAddToTop();
				slCheckForWin();
			}
		}
		return false;
	}
	this.cursorRoutine = function(card) {
		if (this.canDragTopRoutine() && card == this.topCard()) { // easy
			return "pointer";
		}
		if (this.dragPileRule) {
			// handle multiple via rule
			var cardArray = new Array();
			var valueArray = new Array();
			var found  = false;
			for (var i =0; i <this.pile.length;i++) {
				if (this.pile[i] == card) {
					found = true; // starting from here...
					if (!this.pile[i].faceUp) {
						return null; // bail now, we're still face down
					}
				}
				if (found) {
					// add to the list of things we want to drag
					cardArray.push(this.pile[i]);
					valueArray.push(this.pile[i].cardValue);
				}
			}
			if (this.dragPileRule === true || this.dragPileRule(valueArray)) { // true => all face up
				return "pointer";
			}
		}
		return null;
	}
	this.clickRoutine = function(card, x, y) {
		if (this.canDragTopRoutine() && card == this.topCard()) { // easy
			return slStartDrag(new Array(card), x, y);
		}
		if (this.dragPileRule) {
			// handle multiple via rule
			var cardArray = new Array();
			var valueArray = new Array();
			var found  = false;
			for (var i =0; i <this.pile.length;i++) {
				if (this.pile[i] == card) {
					found = true; // starting from here...
					if (!this.pile[i].faceUp) {
						return false; // bail now, we're still face down
					}
				}
				if (found) {
					// add to the list of things we want to drag
					cardArray.push(this.pile[i]);
					valueArray.push(this.pile[i].cardValue);
				}
			}
			if (this.dragPileRule === true || this.dragPileRule(valueArray)) { // true => all face up
				return slStartDrag(cardArray, x, y);
			}
		}
		return false;
	};
	this.isTopOfPile = function(cardDiv) {
		if (this.pile.length == 0) {
			return false; // no cards in pile
		}
		return this.pile[this.pile.length-1] === cardDiv; // last card is the top of the pile
	}
	this.staggerTop = function(cardDiv) {
		if (this.cardXStagger < 0) {
			cardDiv.style.left = this.div.offsetWidth + (this.pile.length - 1) * this.cardXStagger - cardDiv.offsetWidth;
		} else {
			cardDiv.style.left = (this.pile.length - 1) * this.cardXStagger;
		}
		if (this.cardYStagger < 0) {
			cardDiv.style.top = this.div.offsetHeight + (this.pile.length - 1) * this.cardYStagger - cardDiv.offsetHeight;
		} else {
			cardDiv.style.top = (this.pile.length - 1) * this.cardYStagger;
		}
	};
	this.currentEmptyImage = 'Area';
	this.autoFillFromPile = null;
	this.autoFillEmpty = function() {
		if (this.autoFillFromPile && this.autoFillFromPile.topCard() != null) {
			this.addToTop(this.autoFillFromPile.removeTop());  // move from the other deck over to this one
		}
	}
	this.autoDisplayEmpty = function() {
		if (this.pile.length == 0) { // try to autofill the empty thing
			this.autoFillEmpty();
		}
		if (this.pile.length) {
			this.emptyImage.style.display = 'none';
		} else if (this.showsEmptyImage && this.currentEmptyImage != null) {
			if (this.cardXStagger < 0) {
				this.emptyImage.style.left = this.div.offsetWidth - slCardWidth; // flush on the right side
			}
			if (this.cardYStager < 0) {
				this.emptyImage.style.top = this.div.offsetHeight - slCardHeight;
			}
			this.emptyImage.style.background = slMarkerImage(this.currentEmptyImage);
			this.emptyImage.style.display = 'block';
		}
	};
	this.addToTop = function(cardDiv) {
		this.emptyImage.style.display = 'none'; // hide empty image
		cardDiv.style.zIndex = this.baseZIndex + this.pile.length+1; // put it on top of the pile
		if (this.pile.length) { // remove old click handler
			//this.pile[this.pile.length - 1].onmousedown = null;
		}
		this.div.appendChild(cardDiv)
		cardDiv.onmouseover = slCardCursor;
		cardDiv.onmousedown = slCardClicked; // "alert('onMouseDown');";
		cardDiv.pile = this; // so we know how to handle clicking on the card
		this.pile.push(cardDiv); // add to our list
		this.staggerTop(cardDiv); // and stagger it accordingly
		slFly(cardDiv, this.div);
		this.autoDisplayEmpty();
	};
	this.removePairs = false;
	this.postAddToTop = function() { // called after _user_ placed a card here
		if (this.removePairs) {
			var cardDiv = this.removeTop(true);
			slFlipCardDown(cardDiv);
			slGlobalDiscardPile.addToTop(cardDiv);
			cardDiv = this.removeTop();
			slFlipCardDown(cardDiv);
			slGlobalDiscardPile.addToTop(cardDiv);
		}
	};
	this.topCard = function() {
		if (this.pile.length == 0) {
			return null;
		}
		return this.pile[this.pile.length-1];
	};
	this.flipTop = false;
	this.removeTop = function(prepFly) {
		var card = this.pile.pop();
		if (prepFly) {
			slFlyStartX = card.offsetLeft + this.div.offsetLeft;
			slFlyStartY = card.offsetTop + this.div.offsetTop;
		}
		this.autoDisplayEmpty();
		this.div.removeChild(card);
		if (this.pile.length) {
			//this.pile[this.pile.length-1].onmousedown = slCardClicked; // set top card to be clicked
			if (this.flipTop) { // make sure top is up
				slFlipCardUp(this.pile[this.pile.length-1]);
			}
		}
		return card;
	};
	this.clearPile = function() {
		for (var i=0;i<this.pile.length;i++) {
			var card = this.pile[i];
			this.div.removeChild(card)
		}
		this.pile = new Array();
		this.autoDisplayEmpty();
		this.resetPile();
	}
	this.resetPile = function() { } // for things like redeal counts
	this.hitTestDrop = function(x, y) { // hit test is in board coordinates
		// we normally check for the top card
		x -= this.div.offsetLeft; //parseInt(this.div.style.left);
		y -= this.div.offsetTop; //parseInt(this.div.style.top);
		if (this.pile.length > 0) {
			x -= this.topCard().offsetLeft; // parseInt(this.topCard().style.left);
			y -= this.topCard().offsetTop; //parseInt(this.topCard().style.top);
		}
		if (x < 0 || y < 0) {
			return false;
		}
		if (x > slCardWidth || y > slCardHeight) {
			return false;
		}
		return true;
	}
	this.dropRule = true;
	this.blankPileTopCard = "0X";
	this.canDropCardList = function(cardList) {
		//if (cardList.length > 1 && this.canDropLists === false) {
		//	return	false; // can't drop a whole list of them
		//}
		if (this.pileLimit && this.pile.length + cardList.length > this.pileLimit) {
			return false; // too may to fit
		}
		if (this.dropRule === true || this.dropRule === false) {
			return this.dropRule; // hard codeded
		}
		if (this.dropRule == null) {
			return false;
		}
		var topCardValue = this.topCard();
		if (topCardValue == null) topCardValue = this.blankPileTopCard;
		else topCardValue = topCardValue.cardValue
		//alert("test " + cardValue + " on " + topCardValue);
		var valueList = new Array(); // make list of values
		for (var i=0;i<cardList.length;i++) {
			valueList.push(cardList[i].cardValue);
		}
		return this.dropRule(topCardValue, valueList);
	}
	this.highlightDrop = function() {
		var hl = this.topCard();
		if (hl == null) hl = this.emptyImage;
		this.currentHighlight = document.createElement('div');
		hl.appendChild(this.currentHighlight);
		this.currentHighlight.style.left = "0px";
		this.currentHighlight.style.top = "0px";
		this.currentHighlight.style.right = "0px";
		this.currentHighlight.style.bottom = "0px";
		this.currentHighlight.style.position = "absolute";
		this.currentHighlight.style.backgroundColor = 'blue';
		this.currentHighlight.style.opacity = 0.5;
		this.currentHighlight.style.border = "1px solid white";
	}
	this.unhighlightDrop = function() {
		if (this.currentHighlight) {
			// get rid of it
			this.currentHighlight.offsetParent.removeChild(this.currentHighlight);
			this.currentHighlight = null;
		}
	}
	this.isComplete = function() { // complete if pile is empty, by default
		return this.pile.length == 0;
	}
	this.ruleDescription = function() {
		var retval = new Array()
		if (this.dropRule) {
			if (this.dropRule == true) retval.push("Any card can be dropped here.");
			else {
				retval.push(this.dropRule.description);
				if (this.blankPileTopCard.slice(0,1) == '0') {
					retval.push(this.dropRule.blankDesc);
				} else if (this.blankPileTopCard == '*X') {
					retval.push("Spaces can be filled with any card.");
				} else if (this.blankPileTopCard == '--') {
					//retval.push("Spaces can be filled with any card.");
				}
			}
			if (this.pileLimit != 0) {
				if (this.pileLimit == 1) {
					//retval.push("There is a limit of a single card in a pile");
					// handled below
				} else {
					retval.push("There is a limit of " + this.pileLimit + " cards in a pile");
				}
			}
		}
		if (this.autoFillFromPile) {
			retval.push("Spaces are automatically filled from the stock.");
		}
		if (this.canDragTopRoutine()) {
			if (this.dragPileRule == true) {
				retval.push("A sequence of face up cards may be moved as a whole.");
			} else if (this.dragPileRule) {
				retval.push("A sequence of " + this.dragPileRule.description + " may be moved as a whole.");
			} else if (this.pileLimit == 1) {
				retval.push("Only a single card can be played here.");
			} else {
				retval.push("Only the top card is available for play.");
			}
		}
		return retval.join("  ");
	}
	this.custMarshalToString = function () {}
	this.marshalToString = function() {
		var retval = new Array
		for (var i=0;i<this.pile.length;i++) {
			retval += marshalCard(this.pile[i]);
		}
		retval.push(this.custMarshalToString());
		return retval.join(',');
	}
	this.custMarshalFromString = function(str) {}
	this.marshalFromString = function(str) {

	}
}

function slDeckPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.flipToWasteCount = 1; // we flip cards 1 at a time
	this.wastePile = null; // what do we flip to
	this.handPile = null; // unless we've got a hand pile
	this.hasHand = false;
	this.dealLimit = null; // no limit on redeals
	this.dealsLeft = 0;
	this.currentEmptyImage = "Redeal";
	this.dealToPiles = null; // so we can deal onto tableaux piles instead
	this.clickRoutine = function(card, x,y) {
		if (this.handPile) {
			// and only supports 1
			if (this.pile.length == 0) {
				return false; // everything flipped
			}
			if (this.handPile.topCard()) {
				var card = this.handPile.removeTop(); // which will call us to get a new card
				this.wastePile.addToTop(card);
			} else {
				this.handPile.autoFillEmpty();
			}
			return false;
		}
		for (var i=0;i<this.flipToWasteCount;i++) {
			if (this.dealToPiles) {
				for (var j=0;j<this.dealToPiles.length;j++) {
					if (this.pile.length == 0) {
						break; // everything flipped
					}
					// fly converted to table space
					var cardDiv = this.removeTop(true);
					slFlipCardUp(cardDiv);
					this.dealToPiles[j].addToTop(cardDiv);
				}
			} else {
				if (this.pile.length == 0) {
					break; // everything flipped
				}
				// fly converted to table space
				var cardDiv = this.removeTop(true);
				slFlipCardUp(cardDiv);
				this.wastePile.addToTop(cardDiv);
			}
		}
		return false;
	}
	this.cursorRoutine = function(card) {
		if (this.isTopOfPile(card)) return "pointer";
		return null;
	}
	this.dropRule = false;// can't play on deck
	this.showsEmptyImage = true; // has empty image
	this.emptyClickRoutine = function() {
		if (this.dealLimit == null) {
			// unlimited
		} else if (this.dealsLeft == 0) {
			return; // no redeals left
		} else {
			this.dealsLeft -= 1; // one fewer deals
		}
		// re-seed the deck from the waste pile
		if (this.handPile && this.handPile.topCard() != null) {
			var card = this.handPile.removeTop();
			slFlipCardDown(card);
			this.addToTop(card);
		}
		while (this.wastePile.topCard() != null) {
			var card = this.wastePile.removeTop();
			slFlipCardDown(card);
			this.addToTop(card);
		}
		return false;
	}
	this.resetPile = function() {
		this.dealsLeft = this.dealLimit; // reset number of deals
	}
	this.autoFillEmpty = function() {
		// see if there is a waste pile to work from
		if (this.wastePile == null) {
			this.showsEmptyImage = false; // no redeal since no waste
			return;
		}
		if (this.dealLimit != null && this.dealsLeft == 0) {
			this.showsEmptyImage = false; // no redeals left
			return;
		}
		if (this.wastePile.topCard() != null) {
			this.showsEmptyImage = true; // show "deal" if there is something to pull from
		} else {
			// we're now empty
			this.showsEmptyImage = false;
		}
	}
	this.ruleDescription = function() {
		if (this.dealToPiles) {
			return "Deal out a card onto each of the tableaux piles, until the stock is exhausted.";
		}
		var retval = new Array();
		if (this.flipToWasteCount == 1) {
			retval.push("Flip a single card onto the waste pile.");
		} else {
			retval.push("Flip " + this.flipToWasteCount + " cards at a time onto the waste pile.");
		}
		if (this.dealLimit == null) {
			retval.push("Unlimited redeals allowed.");
		} else if (this.dealLimit == 0) {
			retval.push("No redeals allowed.");
		} else {
			retval.push(this.dealLimit + " redeals allowed.");
		}
		return retval.join('  ');
	}
}

function slDeckPileFillEmptyFirst(div) {
	this.inheritsFrom = slDeckPile;
	this.inheritsFrom(div);
	this.superClickRoutine = this.clickRoutine;
	this.clickRoutine = function(card, x,y) {
		// first, see if there is an empty pile in the dealToPiles - if so, do one there
		if (this.dealToPiles) {
			for (var i=0; i<this.dealToPiles.length; i++) {
				if (this.dealToPiles[i].topCard() == null) {
					// deal a single card here
					var cardDiv = this.removeTop(true);
					slFlipCardUp(cardDiv);
					this.dealToPiles[i].addToTop(cardDiv);
					return false;
				}
			}
		}
		// otherwise, call super
		return this.superClickRoutine(card, x, y);
	}
	this.ruleDescription = function() {
		return "If any of the tableaux piles are empty, a single is dealt to it.  Otherwise, deal out a card onto each of the tableaux piles, until the stock is exhausted.";
	}
}

function slWastePile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.cardXStagger = slUpCardsXOverlap;
	this.cardYStagger = slUpCardsYOverlap;
	this.deckPile = null; // what do we go back to
	this.staggerPhase = 3; // how many things to we scatter when we stagger
	this.staggerTop = function(cardDiv) {
		var stagger = (this.pile.length + this.staggerPhase - 1) % this.staggerPhase; // convert length 1 to stagger 0
		cardDiv.style.left = ( stagger * this.cardXStagger) + "px";
		cardDiv.style.top = ( stagger * this.cardYStagger) + "px";
	}
	this.canDragTop = true; // can take top ccard off
	this.dragPileRule = null; // but no pile dragging
	this.dropRule = false; // can't play on waste pile
}

function slHandPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.deckPile = null;
	this.wastePile = null;
	this.dropRule = false;
	this.canDragTop = true;
	this.autoFillEmpty = function() { // draw from deck
		if (this.deckPile.topCard() != null) {
			var card = this.deckPile.removeTop(true);
			slFlipCardUp(card);
			this.addToTop(card);
		}
	}
	this.postAddToTop = function() { // called after _user_ placed a card here (which only happens if we add drop rule)
		var cardDiv = this.removeTop(true);
		slFlipCardDown(cardDiv);
		slGlobalDiscardPile.addToTop(cardDiv);
		cardDiv = this.removeTop();
		slFlipCardDown(cardDiv);
		slGlobalDiscardPile.addToTop(cardDiv);
	};
	this.isComplete = function() {
		return this.pile.length == 0; // we aren't complete if there is a card here
	}
}

slGlobalDiscardPile = null;
function slDiscardPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.cardXStagger = 3;
	this.cardYStagger = 0;
	this.deckPile = null; // what do we go back to
	this.staggerPhase = 0; // how many things to we scatter when we stagger
	slGlobalDiscardPile = this; // only one of these globally
	this.canDragTop = false; // can't drag to here
	this.dragPileRule = null; // but no pile dragging
	this.dropRule = null;
	this.showsEmptyImage = true;
	this.ruleDescription = function() {
		return "Cards are discarded here.";
	}
	this.isComplete = function() { // complete if pile is all cards
		return this.pile.length == slNumDecks * 52;
	}
}
function slDiscardPileKings(div) {
	this.inheritsFrom = slDiscardPile;
	this.inheritsFrom(div);
	this.dropRule = slDropKingsOnly;
	this.postAddToTop = function() { // called after _user_ placed a card here
		slFlipCardDown(this.topCard());
	};
}

function slTableauxPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.cardYStagger = slUpCardsYOverlap;
	this.cardXStagger = 0; // build down by default
	this.downCardYStagger = slDownCardsYOverlap;
	this.downCardXStagger = 0;
	this.canDragTop = true;
	this.dropRule = slDropRuleAltColorDown; // default to alternating
	this.dragPileRule = slDragRuleAltColorDown; // and can drag a pile
	this.flipTop = true; // and flip the next card
	this.showsEmptyImage = true;
	this.staggerTop = function(cardDiv) {
		// add to the bottom
		var top = 0 * parseInt(this.div.style.top);
		var left = 0 * parseInt(this.div.style.left);
		for (var i=0;i<this.pile.length-1;i++) {
			if (this.pile[i].faceUp) {
				top += this.cardYStagger;
				left += this.cardXStagger;
			} else {
				top += this.downCardYStagger;
				left += this.downCardXStagger;
			}

		}
		cardDiv.style.top = top + "px";
		cardDiv.style.left = left + "px";
	}
	this.bestDropRank = 50; // an OK thing to auto-drop
}


function slFoundationPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	//this.pileLimit = 13; // we assume when we fill it up, we're done
	this.canDragTop = true; // can drag out from foundation
	this.blankPileTopCard = "0X";
	this.dropRule = slDropRuleFoundation;
	this.isComplete = function() {
		return this.pile.length == 13; // got all the cards
	}
	this.currentEmptyImage = 'Foundation';
	this.showsEmptyImage = true;
	this.bestDropRank = 999; // best place to auto-drop something
}
function slFoundationPileClubs(div) {
	this.inheritsFrom = slFoundationPile;
	this.inheritsFrom(div);
	this.blankPileTopCard = "0C";
}
function slFoundationPileHearts(div) {
	this.inheritsFrom = slFoundationPile;
	this.inheritsFrom(div);
	this.blankPileTopCard = "0H";
}
function slFoundationPileDiamonds(div) {
	this.inheritsFrom = slFoundationPile;
	this.inheritsFrom(div);
	this.blankPileTopCard = "0D";
}
function slFoundationPileSpades(div) {
	this.inheritsFrom = slFoundationPile;
	this.inheritsFrom(div);
	this.blankPileTopCard = "0S";
}

function slReservePile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.canDragTop = true; // can take top ccard off
	this.dragPileRule = null; // but no pile dragging
	this.dropRule = false; // can't play on waste pile
	this.cardYStagger = slUpCardsYOverlap;
}

function slFreeCellPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.currentEmptyImage = 'Freecell';
	this.canDragTop = true; // can take top ccard off
	this.dragPileRule = null; // but no pile dragging
	this.dropRule = true; // can drop anything here
	this.pileLimit = 1; // but only one
	this.showsEmptyImage = true;
	this.isComplete = function() {
		return this.pile.length == 0; // we aren't complete if there is a card here
	}
}

function slPickAnyOne(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.canDragTop = true; // can take top ccard off
	this.dragPileRule = null; // but no pile dragging
	this.dropRule = false; // can drop anything here
	this.pileLimit = 1; // but only one
	this.showsEmptyImage = false;
	this.isComplete = function() {
		return this.pile.length == 0; // we aren't complete if there is a card here
	}
	this.clickRoutine = function(card, x, y) { // we can drag anything
		this.willRemoveCard = card;
		return slStartDrag(new Array(card), x, y);
	};
	this.removeTop = function() {
		// since we aren't really removing the top, rather one specific thing...
		if (!this.willRemoveCard) {
			return null;
		}
		for (var i=0;i<this.pile.length;i++) {
			if (this.pile[i] == this.willRemoveCard) {
				// remove this one
				this.pile.splice(i,1);
				break;
			}
		}
		var card = this.willRemoveCard;
		this.div.removeChild(this.willRemoveCard);
		this.willRemoveCard = null;
		return card;
	};
	this.ruleDescription = function() {
		return "Any card from the deck can be picked and played, regardless of its location in the pile.";
	};
}

// for "drag and combine cards to discard" games
function slCombinePairsPile(div) {
	this.inheritsFrom = slPileBase;
	this.inheritsFrom(div);
	this.canDragTop = true; // can take top card off
	this.dragPileRule = false;// but no pile dragging
	this.dropRule = null; // needs to be explicitly set
	this.postAddToTop = function() { // called after _user_ placed a card here
		// user dropped a card here, so it must total 14, so remove them
		var cardDiv = this.removeTop(true);
		slFlipCardDown(cardDiv);
		slGlobalDiscardPile.addToTop(cardDiv);
		cardDiv = this.removeTop();
		slFlipCardDown(cardDiv);
		slGlobalDiscardPile.addToTop(cardDiv);
	};
	this.isComplete = function() {
		return this.pile.length == 0; // we aren't complete if there is a card here
	}
}
function slCombinePairs14Pile(div) {
	this.inheritsFrom = slCombinePairsPile;
	this.inheritsFrom(div);
	this.dropRule = slDropRuleTotal14;
}
function slCombinePairs13Pile(div) {
	this.inheritsFrom = slCombinePairsPile;
	this.inheritsFrom(div);
	this.dropRule = slDropRuleTotal13;
}

// similar to above, but for pyramid style games
function slStaggerOverlapPile(div) {
	this.inheritsFrom = slCombinePairsPile;
	this.inheritsFrom(div);
	// can take top card off?
	this.leftPile = null;
	this.rightPile = null;
	this.canDragTopRoutine = function() {
		// check left and right to see if they are empty
		var left = true;
		if (this.leftPile) {
			left = this.leftPile.isComplete();
		}
		var right = true;
		if (this.rightPile) {
			right = this.rightPile.isComplete();
		}
		return left && right;
	}
	this.superCanDropCardList = this.canDropCardList;
	this.canDropCardList = function(cardList) {
		if (!this.canDragTopRoutine()) {
			return false; // we're blocked
		}
		return this.superCanDropCardList.call(this,cardList);
	}
}
function slStaggerOverlap14Pile(div) {
	this.inheritsFrom = slStaggerOverlapPile;
	this.inheritsFrom(div);
	this.dropRule = slDropRuleTotal14;
}
function slStaggerOverlap13Pile(div) {
	this.inheritsFrom = slStaggerOverlapPile;
	this.inheritsFrom(div);
	this.dropRule = slDropRuleTotal13;
}
function slStaggerOverlapPairsPile(div) {
	this.inheritsFrom = slStaggerOverlapPile;
	this.inheritsFrom(div);
	this.dropRule = slDropRulePairs;
}


// the various card building rules

function slCreatePile(x,y,width,height,kind, name)
{
	var div = document.createElement('div');
	div.setAttribute('class','slPile');
	div.setAttribute('id', name);
	div.style.position = 'absolute';
	if (x < 0) {
		div.style.right = (-x) + 'px';
	} else {
		div.style.left = x + 'px';
	}
	if (y < 0) {
		div.style.bottom = (-y) + 'px';
	} else {
		div.style.top = y + 'px';
	}
	div.style.width = width + 'px';
	div.style.height = height + 'px';
	slCardTable.appendChild(div);
	var pile = new kind(div);
	return pile;
}

function slClearAllPiles()
{
	if (slWinTimer) {
		clearInterval(slWinTimer);
		slWinTimer = null;
	}
	if (slFlyTimer) {
		clearInterval(slFlyTimer);
		slFlyTimer = null;
	}

	for (var i=0;i<slAllPiles.length;i++) {
		slAllPiles[i].clearPile();
	}
}

function slRebrandCards()
{
	for (var i=0;i<slAllPiles.length;i++) {
		for (var j=0;j<slAllPiles[i].pile.length;j++) {
			slRebrandCard(slAllPiles[i].pile[j]);
		}
	}
}

function slInit()
{
	slCardTable = document.getElementById('slcardtable');
	slCardTableWidth = slCardTable.offsetWidth;
	slCardTableHeight = slCardTable.offsetHeight;
}

function slReset()
{
	if (slWinTimer) {
		clearInterval(slWinTimer);
		slWinTimer = null;
	}
	for (var i=0;i<slAllPiles.length;i++) {
		slAllPiles[i].clearPile();
		slCardTable.removeChild(slAllPiles[i].div); // remove this div
	}
	slAllPiles = new Array();
	slFlyEnabled = false;
	if (slFlyQueue) {
		clearInterval(slFlyQueue);
		slFlyQueue = null;
	}
	slFlyQueue = new Array();
}

function slAnimateWin()
{
	if (slFlyTimer) return; //	don't animate while still flying the last cards
	for (var i=0;i<slAllPiles.length;i++) {
		for (var j=0;j<slAllPiles[i].pile.length;j++) {
			var cardDiv = slAllPiles[i].pile[j];
			var x = parseInt(cardDiv.style.left);
			var y = parseInt(cardDiv.style.top);
			if (!cardDiv.winDX) {
				cardDiv.winDX = Math.floor(Math.random() * 5) - 2;
			}
			if (!cardDiv.winDY) {
				cardDiv.winDY = Math.floor(Math.random() * 5) - 2;
			}
			// offset by our amount
			x += cardDiv.winDX;
			y += cardDiv.winDY;
			x += slAllPiles[i].div.offsetLeft; // convert to board coord
			y += slAllPiles[i].div.offsetTop;
			if (x < 0) { x = 0; cardDiv.winDX = Math.floor(Math.random() * 5) - 2; }
			if (y < 0) { y = 0; cardDiv.winDY = Math.floor(Math.random() * 5) - 2; }
			if (x > slCardTable.offsetWidth - cardDiv.offsetWidth) {
				x = slCardTable.offsetWidth - cardDiv.offsetWidth;
				cardDiv.winDX = Math.floor(Math.random() * 5) - 2;
			}
			if (y > slCardTable.offsetHeight - cardDiv.offsetHeight) {
				y = slCardTable.offsetHeight - cardDiv.offsetHeight;
				cardDiv.winDY = Math.floor(Math.random() * 5) - 2;
			}
			// back to pile coord
			x -= slAllPiles[i].div.offsetLeft;
			y -= slAllPiles[i].div.offsetTop;
			cardDiv.style.left = x + "px";
			cardDiv.style.top = y + "px";
		}
	}
}
slTurnCallback = function() {}
slWonCallback = function() {}

function slCheckForWin()
{
	for (var i=0;i<slAllPiles.length;i++) {
		if (!slAllPiles[i].isComplete()) {
			slTurnCallback();
			return false;
		}
	}
	alert('Game over');
	slWonCallback();
	slWinTimer = setInterval("slAnimateWin();",100);
	return true;
}